@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -43,6 +43,8 @@
 #undef HAVE_POLLIN
 #undef HAVE_SS_SP
 #undef HAVE_SS_BASE
+#undef HAVE_LONGLONG
+#undef HAVE_LONGDOUBLE
 #undef PTH_DEBUG
 #undef PTH_NSIG
 #undef PTH_MCTX_MTH_use
