@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999-2001 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -114,6 +114,8 @@ pth_event_t pth_event(unsigned long spec, ...)
     if (spec & PTH_EVENT_FD) {
         /* filedescriptor event */
         int fd = va_arg(ap, int);
+        if (!pth_util_fd_valid(fd))
+            return_errno(NULL, EBADF);
         ev->ev_type = PTH_EVENT_FD;
         ev->ev_goal = (int)(spec & (PTH_UNTIL_FD_READABLE|\
                                     PTH_UNTIL_FD_WRITEABLE|\
@@ -413,8 +415,10 @@ int pth_wait(pth_event_t ev_ring)
     ev = ev_ring;
     occurred = 0;
     do {
-        occurred++;
-        pth_debug2("pth_wait: occurred event 0x%lx", (unsigned long)ev);
+        if (ev->ev_occurred) {
+            pth_debug2("pth_wait: occurred event 0x%lx", (unsigned long)ev);
+            occurred++;
+        }
         ev = ev->ev_next;
     } while (ev != ev_ring);
 
