@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999-2001 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -290,7 +290,7 @@ int pth_select_ev(int nfd, fd_set *rfds, fd_set *wfds,
     /* now directly poll filedescriptor sets to avoid unneccessary
        (and resource consuming because of context switches, etc) event
        handling through the scheduler. We've to be carefully here, because not
-       all platforms guarranty us that the sets are unmodified when an error
+       all platforms guarranty us that the sets are unmodified if an error
        or timeout occured. */
     delay.tv_sec  = 0;
     delay.tv_usec = 0;
@@ -411,8 +411,8 @@ int pth_poll_ev(struct pollfd *pfd, nfds_t nfd, int timeout, pth_event_t ev_extr
     FD_ZERO(&wfds);
     FD_ZERO(&efds);
     for(i = 0; i < nfd; i++) {
-        if (pfd[i].fd < 0)
-            continue;
+        if (!pth_util_fd_valid(pfd[i].fd))
+            return_errno(-1, EBADF);
         if (pfd[i].events & POLLIN)
             FD_SET(pfd[i].fd, &rfds);
         if (pfd[i].events & POLLOUT)
@@ -495,8 +495,8 @@ int pth_connect_ev(int s, const struct sockaddr *addr, socklen_t addrlen, pth_ev
     /* restore filedescriptor mode */
     errno_shield { pth_fdmode(s, fdmode); }
 
-    /* when it is still on progress wait until socket is really writeable */
-    if (rv == -1 && errno == EINPROGRESS) {
+    /* if it is still on progress wait until socket is really writeable */
+    if (rv == -1 && errno == EINPROGRESS && fdmode != PTH_FDMODE_NONBLOCK) {
         ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, s);
         if (ev_extra != NULL)
             pth_event_concat(ev, ev_extra, NULL);
@@ -541,7 +541,8 @@ int pth_accept_ev(int s, struct sockaddr *addr, socklen_t *addrlen, pth_event_t
     /* poll socket via accept */
     ev = NULL;
     while ((rv = pth_sc(accept)(s, addr, addrlen)) == -1
-           && (errno == EAGAIN || errno == EWOULDBLOCK)) {
+           && (errno == EAGAIN || errno == EWOULDBLOCK)
+           && fdmode != PTH_FDMODE_NONBLOCK) {
         /* do lazy event allocation */
         if (ev == NULL) {
             ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, s);
@@ -593,12 +594,14 @@ ssize_t pth_read_ev(int fd, void *buf, size_t nbytes, pth_event_t ev_extra)
     if (nbytes == 0)
         return 0;
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (pth_fdmode(fd, PTH_FDMODE_POLL) == PTH_FDMODE_BLOCK) {
 
         /* now directly poll filedescriptor for readability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd))
+            return_errno(-1, EBADF);
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -606,7 +609,7 @@ ssize_t pth_read_ev(int fd, void *buf, size_t nbytes, pth_event_t ev_extra)
         while ((n = pth_sc(select)(fd+1, &fds, NULL, NULL, &delay)) < 0
                && errno == EINTR) ;
 
-        /* when filedescriptor is still not readable,
+        /* if filedescriptor is still not readable,
            let thread sleep until it is or the extra event occurs */
         if (n < 1) {
             ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -661,12 +664,16 @@ ssize_t pth_write_ev(int fd, const void *buf, size_t nbytes, pth_event_t ev_extr
     /* force filedescriptor into non-blocking mode */
     fdmode = pth_fdmode(fd, PTH_FDMODE_NONBLOCK);
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (fdmode != PTH_FDMODE_NONBLOCK) {
 
         /* now directly poll filedescriptor for writeability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd)) {
+            pth_fdmode(fd, fdmode);
+            return_errno(-1, EBADF);
+        }
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -676,7 +683,7 @@ ssize_t pth_write_ev(int fd, const void *buf, size_t nbytes, pth_event_t ev_extr
 
         rv = 0;
         for (;;) {
-            /* when filedescriptor is still not writeable,
+            /* if filedescriptor is still not writeable,
                let thread sleep until it is or event occurs */
             if (n < 1) {
                 ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -717,7 +724,7 @@ ssize_t pth_write_ev(int fd, const void *buf, size_t nbytes, pth_event_t ev_extr
         }
     }
     else {
-        /* Now perform the actual write operation */
+        /* just perform the actual write operation */
         while ((rv = pth_sc(write)(fd, buf, nbytes)) < 0
                && errno == EINTR) ;
     }
@@ -751,12 +758,14 @@ ssize_t pth_readv_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t ev
     if (iovcnt <= 0 || iovcnt > UIO_MAXIOV)
         return_errno(-1, EINVAL);
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (pth_fdmode(fd, PTH_FDMODE_POLL) == PTH_FDMODE_BLOCK) {
 
         /* first directly poll filedescriptor for readability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd))
+            return_errno(-1, EBADF);
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -764,7 +773,7 @@ ssize_t pth_readv_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t ev
         while ((n = pth_sc(select)(fd+1, &fds, NULL, NULL, &delay)) < 0
                && errno == EINTR) ;
 
-        /* when filedescriptor is still not readable,
+        /* if filedescriptor is still not readable,
            let thread sleep until it is or event occurs */
         if (n < 1) {
             ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -860,6 +869,9 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
     ssize_t rv;
     ssize_t s;
     int n;
+    struct iovec tiov_stack[32];
+    struct iovec *tiov;
+    int tiovcnt;
 
     pth_implicit_init();
     pth_debug2("pth_writev_ev: enter from thread \"%s\"", pth_current->name);
@@ -871,8 +883,18 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
     /* force filedescriptor into non-blocking mode */
     fdmode = pth_fdmode(fd, PTH_FDMODE_NONBLOCK);
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (fdmode != PTH_FDMODE_NONBLOCK) {
+        /* provide temporary iovec structure */
+        if (iovcnt > sizeof(tiov_stack)) {
+            tiovcnt = (sizeof(struct iovec) * UIO_MAXIOV);
+            if ((tiov = (struct iovec *)malloc(tiovcnt)) == NULL)
+                return -1 /* errno is set */;
+        }
+        else {
+            tiovcnt = sizeof(tiov_stack);
+            tiov    = tiov_stack;
+        }
 
         /* init return value and number of bytes to write */
         rv      = 0;
@@ -881,11 +903,17 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
         /* init local iovec structure */
         liov    = NULL;
         liovcnt = 0;
-        pth_writev_iov_advance(iov, iovcnt, 0, &liov, &liovcnt);
+        pth_writev_iov_advance(iov, iovcnt, 0, &liov, &liovcnt, tiov, tiovcnt);
 
         /* first directly poll filedescriptor for writeability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd)) {
+            pth_fdmode(fd, fdmode);
+            if (iovcnt > sizeof(tiov_stack))
+                free(tiov);
+            return_errno(-1, EBADF);
+        }
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -894,7 +922,7 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
                && errno == EINTR) ;
 
         for (;;) {
-            /* when filedescriptor is still not writeable,
+            /* if filedescriptor is still not writeable,
                let thread sleep until it is or event occurs */
             if (n < 1) {
                 ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -905,12 +933,14 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
                     pth_event_isolate(ev);
                     if (!pth_event_occurred(ev)) {
                         pth_fdmode(fd, fdmode);
+                        if (iovcnt > sizeof(tiov_stack))
+                            free(tiov);
                         return_errno(-1, EINTR);
                     }
                 }
             }
 
-            /* Now perform the actual write operation */
+            /* now perform the actual write operation */
 #if PTH_FAKE_RWV
             while ((s = pth_writev_faked(fd, liov, liovcnt)) < 0
                    && errno == EINTR) ;
@@ -926,7 +956,7 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
                we've to mimic the usual blocking I/O behaviour of writev(2) */
             if (s > 0 && s < (ssize_t)nbytes) {
                 nbytes -= s;
-                pth_writev_iov_advance(iov, iovcnt, n, &liov, &liovcnt);
+                pth_writev_iov_advance(iov, iovcnt, s, &liov, &liovcnt, tiov, tiovcnt);
                 n = 0;
                 continue;
             }
@@ -938,9 +968,13 @@ ssize_t pth_writev_ev(int fd, const struct iovec *iov, int iovcnt, pth_event_t e
             /* stop looping */
             break;
         }
+
+        /* cleanup */
+        if (iovcnt > sizeof(tiov_stack))
+            free(tiov);
     }
     else {
-        /* Now perform the actual write operation */
+        /* just perform the actual write operation */
 #if PTH_FAKE_RWV
         while ((rv = pth_writev_faked(fd, iov, iovcnt)) < 0
                && errno == EINTR) ;
@@ -974,9 +1008,9 @@ intern ssize_t pth_writev_iov_bytes(const struct iovec *iov, int iovcnt)
 
 /* advance the virtual pointer of a struct iov */
 intern void pth_writev_iov_advance(const struct iovec *riov, int riovcnt, size_t advance,
-                                   struct iovec **liov, int *liovcnt)
+                                   struct iovec **liov, int *liovcnt, 
+                                   struct iovec *tiov, int tiovcnt)
 {
-    static struct iovec siov[UIO_MAXIOV];
     int i;
 
     if (*liov == NULL && *liovcnt == 0) {
@@ -987,10 +1021,10 @@ intern void pth_writev_iov_advance(const struct iovec *riov, int riovcnt, size_t
     if (advance > 0) {
         if (*liov == riov && *liovcnt == riovcnt) {
             /* reinitialize with a copy to be able to adjust it */
-            *liov = &siov[0];
+            *liov = &tiov[0];
             for (i = 0; i < riovcnt; i++) {
-                siov[i].iov_base = riov[i].iov_base;
-                siov[i].iov_len  = riov[i].iov_len;
+                tiov[i].iov_base = riov[i].iov_base;
+                tiov[i].iov_len  = riov[i].iov_len;
             }
         }
         /* advance the virtual pointer */
@@ -1151,12 +1185,14 @@ ssize_t pth_recvfrom_ev(int fd, void *buf, size_t nbytes, int flags, struct sock
     if (nbytes == 0)
         return 0;
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (pth_fdmode(fd, PTH_FDMODE_POLL) == PTH_FDMODE_BLOCK) {
 
         /* now directly poll filedescriptor for readability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd))
+            return_errno(-1, EBADF);
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -1164,7 +1200,7 @@ ssize_t pth_recvfrom_ev(int fd, void *buf, size_t nbytes, int flags, struct sock
         while ((n = pth_sc(select)(fd+1, &fds, NULL, NULL, &delay)) < 0
                && errno == EINTR) ;
 
-        /* when filedescriptor is still not readable,
+        /* if filedescriptor is still not readable,
            let thread sleep until it is or the extra event occurs */
         if (n < 1) {
             ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -1179,7 +1215,7 @@ ssize_t pth_recvfrom_ev(int fd, void *buf, size_t nbytes, int flags, struct sock
         }
     }
 
-    /* Now perform the actual read. We're now guarrantied to not block,
+    /* now perform the actual read. We're now guarrantied to not block,
        either because we were already in non-blocking mode or we determined
        above by polling that the next recvfrom(2) call will not block.  But keep
        in mind, that only 1 next recvfrom(2) call is guarrantied to not block
@@ -1231,12 +1267,16 @@ ssize_t pth_sendto_ev(int fd, const void *buf, size_t nbytes, int flags, const s
     /* force filedescriptor into non-blocking mode */
     fdmode = pth_fdmode(fd, PTH_FDMODE_NONBLOCK);
 
-    /* poll filedescriptor when not already in non-blocking operation */
+    /* poll filedescriptor if not already in non-blocking operation */
     if (fdmode != PTH_FDMODE_NONBLOCK) {
 
         /* now directly poll filedescriptor for writeability
            to avoid unneccessary (and resource consuming because of context
            switches, etc) event handling through the scheduler */
+        if (!pth_util_fd_valid(fd)) {
+            pth_fdmode(fd, fdmode);
+            return_errno(-1, EBADF);
+        }
         FD_ZERO(&fds);
         FD_SET(fd, &fds);
         delay.tv_sec  = 0;
@@ -1246,7 +1286,7 @@ ssize_t pth_sendto_ev(int fd, const void *buf, size_t nbytes, int flags, const s
 
         rv = 0;
         for (;;) {
-            /* when filedescriptor is still not writeable,
+            /* if filedescriptor is still not writeable,
                let thread sleep until it is or event occurs */
             if (n < 1) {
                 ev = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE|PTH_MODE_STATIC, &ev_key, fd);
@@ -1287,7 +1327,7 @@ ssize_t pth_sendto_ev(int fd, const void *buf, size_t nbytes, int flags, const s
         }
     }
     else {
-        /* Now perform the actual send operation */
+        /* just perform the actual send operation */
         while ((rv = pth_sc(sendto)(fd, buf, nbytes, flags, to, tolen)) < 0
                && errno == EINTR) ;
     }
