@@ -1,6 +1,6 @@
 /*
 **  GNU Pth - The GNU Portable Threads
-**  Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
 **
 **  This file is part of GNU Pth, a non-preemptive thread scheduling
 **  library which can be found at http://www.gnu.org/software/pth/.
@@ -149,7 +149,7 @@ intern void pth_time_mul(pth_time_t *t1, int n)
     t1->tv_sec  *= n;
     t1->tv_usec *= n;
     t1->tv_sec  += (t1->tv_usec / 1000000);
-    t1->tv_usec  = (t1->tv_usec & 1000000);
+    t1->tv_usec  = (t1->tv_usec % 1000000);
     return;
 }
 
